from dataclasses import asdict

import datetime
from functools import wraps

from sqlalchemy import MetaData
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import FlushError
from sqlalchemy.orm.session import sessionmaker, Session

Datetime = datetime.datetime


class _Empty: ...

_Session = sessionmaker()
def operation(func):
    @wraps(func)
    def wrapper(instance: MetaData, item, session=None):
        instance.ensure_table_exists(item)

        commit = session is None
        session: Session = session or Session(bind=instance.bind)
        ret = func(instance, item, session)
        if commit:
            try:
                session.commit()
            except (IntegrityError) as ex:
                session.rollback()
                v = session.query(type(item)).get(item.item_id)
                if v != item:
                    diff = asdict(item).items() ^ asdict(v).items()
                    if v.run_date == item.run_date:
                        print(f"Update suppressed: {v.id} {diff}")
                        return
                    raise NotImplementedError(f"Update needed {diff}")
            except FlushError as ex:
                session.rollback()
                session.merge(item)
                session.commit()
            finally:
                session.close()
        if ret == _Empty:
            return None
        return ret or instance

    return wrapper



import datetime
import inspect
import logging
from dataclasses import dataclass, is_dataclass, fields, Field, field
from functools import wraps, partial
from itertools import chain
from typing import Type, Union, Text, List, get_args, get_origin, Any

from sqlalchemy import Column, MetaData, Table, String, create_engine, ForeignKey, event
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import sessionmaker, Session, relationship, ColumnProperty, mapper
from sqlalchemy.orm.attributes import History, HISTORY_BLANK
from sqlalchemy.orm.base import manager_of_class, INIT_OK, PASSIVE_NO_RESULT
from sqlalchemy.sql.sqltypes import NullType, DateTime, Float, JSON, Integer
from sqlalchemy.sql.type_api import TypeEngine

from .proxy import CallbackProxy
from .typemap import python_to_alechamy
from .util import operation
logger = logging.getLogger("mortar")

Datetime = datetime.datetime
class mortar(MetaData):
    tables_checked = []

    def __call__(self, item, *a, **kwargs):
        try:
            return item.__table__
        except:
            return self.defer_to_session(item)

    def engine(self, engine_or_string, **kw):
        if isinstance(engine_or_string, str):
            engine_or_string = create_engine(engine_or_string, **kw)
        self.bind = engine_or_string
        return self

    def ensure_table_exists(self, op):
        if not is_dataclass(op):
            return None
        tbl = op.__table__
        if tbl.name in self.tables_checked:
            return True
        if self.bind.has_table(tbl.name):
            self.tables_checked.append(tbl.name)
            return True
        else:
            self.create_all(tables=[tbl], checkfirst=False)
            self.tables_checked.append(tbl.name)
            return True

    @operation
    def save(self, instance, session=None):
        return session.add_all([instance])

    @operation
    def truncate(self, instance, session=None):
        session.execute(instance.__table__.delete())

    def __getattr__(self, item):
        if item in Session.__dict__:
            return self.defer_to_session(item)

    def defer_to_session(self, item):
        @operation
        def runner(s, o, ses):
            return getattr(ses, item)(o)

        return partial(runner, self)


metadata: Union[mortar, Session] = mortar()


def persist(cls=None, /, *, table=None, database=None) -> Union[Type]:
    closure_table = table

    def wrapper(cls):
        table = closure_table or Table(cls.__name__, database or metadata)
        return _process_class(cls, table)

    if cls is None:
        # parameters sent return the wrapper to decorate
        return wrapper

    # called with nothing as @persist we'll do everthing, return a processed class
    return wrapper(cls)


class _Empty: ...


class Getter(hybrid_property):
    _mortar_prop = True
    accepts_scalar_loader = False
    collection = None
    column_data = None
    supports_population = False

    @property
    def impl(self):
        return self

    def get(self, state, dict_, passive=None):
        return self.fget(state.object)

    def __get__(self, instance, owner):
        if instance is None:
            return self

        return self.fget(instance)

    def __init__(self, descriptor):
        super().__init__(descriptor)
        self.key = descriptor.__name__

    def get_history(self, state, dict_, passive=False):
        if self.key in dict_:
            return History.from_object_attribute(self, state, dict_[self.key])
        else:
            if passive & INIT_OK:
                passive ^= INIT_OK
            current = self.get(state, dict_, passive=passive)
            if current is PASSIVE_NO_RESULT:
                return HISTORY_BLANK
            else:
                return History.from_object_attribute(self, state, current)

class presitable_property(hybrid_property):
    _mortar_prop = True
    column_data = None


Base = declarative_base()


def persist_property(prp_or_column) -> Union[property, Any]:
    prp = None
    clm = None
    if isinstance(prp_or_column, Column):
        clm = prp_or_column
    else:
        prp = prp_or_column

    def wrapper(prp):
        p = Getter(prp)
        p._mortar_prop = True
        p.column_data = clm
        return p

    if prp is None:
        return wrapper

    return wrapper(prp)


def persisted_properties(datacls):
    for k, v in inspect.getmembers(datacls, lambda a: hasattr(a, "_mortar_prop")):
        # Everythiing is set up to use fields from dataclasses so sending a field here makes seense
        f = field()
        f.name = k
        f.type = Text
        f.metadata = {"_mortar_prop": True}
        if v.column_data is not None:
            f.metadata["mortar"] = {"column": v.column_data}

        # setattr(datacls, k, QueryableAttribute(datacls, k, v))
        yield f


def _before_insert(mapper, connection, target):
    target.__dict__["item_id"] = type(target).item_id.fget(target)


def _process_class(datacls, table: Table):
    if not is_dataclass(datacls):
        logging.getLogger("damp").warning(
            f"{datacls} wrapped with persist but is not a data class, making it a data class first"
        )
        datacls = dataclass(datacls)
    props = {}
    manual_register = []
    insert_listener = event.listen(datacls, 'before_insert', _before_insert)

    @event.listens_for(datacls, 'load')
    def enable_persist(target, context):
        target.__dict__["item_id"] = CallbackProxy(lambda: type(target).item_id.fget(target))

    for dt_field in chain(fields(datacls), persisted_properties(datacls)):
        field_type = python_to_alechamy(dt_field.type)
        if is_dataclass(field_type):
            keys = []
            for k in field_type.__table__.primary_key.columns:
                fkc = Column(dt_field.name + "_" + k.key, k.type, ForeignKey(k))
                keys.append(fkc)
                table.append_column(fkc)
            props[dt_field.name] = relationship(field_type, foreign_keys=keys)
        else:
            meta = mortar_meta(dt_field)
            column_def = column_factory(meta, dt_field, field_type)
            table.append_column(column_def)
            if meta.get("_mortar_prop"):
                manual_register.append(column_def.name)
                props = {column_def.name: ColumnProperty(column_def, _instrument=False)}
    datacls.save = lambda s, **kw: table.metadata.save(s, **kw)
    datacls.truncate = classmethod(table.metadata.truncate)
    if len(table.primary_key) == 0:
        pk = table.columns.get(f"item_id", Column('auto_id', String))
        pk.instrument = False
        pk.primary_key = True
        table.append_constraint(pk)
    datacls.__table__ = table
    datacls.__mapper__ = mapper(datacls, table, properties=props)
    manager = manager_of_class(datacls)

    for k in manual_register:
        manager.instrument_attribute(k, getattr(datacls, k))
    return datacls


def mortar_meta(dt_field: Field):
    meta = dt_field.metadata.copy()
    mort_meta = {}
    col_def = None
    if isinstance(meta, Column):
        col_def = meta
    else:
        mort_meta = meta.pop("mortar", mort_meta)
        for k in meta:
            if k[:7] == "_mortar":
                mort_meta[k] = meta[k]
    if col_def is not None:
        mort_meta["column"] = col_def
    return mort_meta


def column_factory(meta: dict, dt_field, field_type=None):
    column = meta.get("column", Column(dt_field.name))
    if not isinstance(column, Column):
        return column
    if column.name is None:
        column.name = dt_field.name
    if type(column.type) is NullType:
        logger.debug((dt_field.name, dt_field.type))
        column.type = (field_type or python_to_alechamy(dt_field.type))

    try:
        return column.copy()
    except:
        return column



from itertools import chain
import uuid
from operator import __all__ as all_op
from operator import *

import weakref

_inplace = (iadd, iand, iconcat, ifloordiv, ilshift, imod, imul,
            imatmul, ior, ipow, irshift, isub, itruediv, ixor)
_python_types = (float, int, list, dict, str)

imported_operators = locals()

__doc__ = """YAP - Yet another Proxy. Provides a complete proxy solution without using sometimes unsafe (and thus restricted) methods like exec. 

YAP is composed of 2 components the registery that holds all proxied items and the proxy implementation. 
"""

class _ProxyRegistery(object):
    """This Registery allows for configuration on how the proxy functions with regards to the proxied value. The registery also provides inflection points to allow for events in the future. """
    
    def __init__(self):
        self._proxied = {}

    def _default_config(self):
        return dict(load_once=False,
                    unpack=True
                    )

    def unpack(self, proxy, *a, **kw):
        key = self.key(proxy)
        proxied, cfg = self._proxied[key]
        unpacked = proxied(*a, **kw)
        if cfg["load_once"]:
            cfg["unpack"] = False
            self._proxied[key] = unpacked, cfg
        return unpacked

    def key(self, proxy):
        return id(proxy)

    def proxy(self, proxy, proxied, config=None, **kw):
        key = self.key(proxy)
        proxy_config = self._default_config()
        proxy_config.update(**kw)
        proxy_config.update(config or {})
        self._proxied[key] = (proxied, proxy_config)

    def get(self, proxy):
        key = self.key(proxy)
        proxied, cfg = self._proxied[key]
        if cfg["unpack"]:
            return self.unpack(proxy)
        return self._proxied[key][0]


proxy_cache = _ProxyRegistery()


def proxied_call(proxy, fn, *a, **kw):
    subject = proxy_cache.get(proxy)
    subject_loc = id(subject)
    if isinstance(fn, str):
        fn = getattr(subject, fn)
        outcome = fn(*a, **kw)
    else:
        outcome = fn(subject, *a, *kw)
    o_id = id(outcome)
    l_id = id(subject)
    if fn in _inplace:
        proxy_cache.proxy(proxy, outcome, unpack=False)
        return proxy
    return outcome


class BasicProxy(object):

    def __repr__(self):
        subject = proxy_cache.get(self)
        return subject.__repr__()

    def __len__(self):
        subject = proxy_cache.get(self)
        return len(subject)

    def __bool__(self):
        subject = proxy_cache.get(self)
        return bool(subject)

    def __setattr__(self, attr, val):
        subject = proxy_cache.get(self)
        setattr(subject, attr, val)

    def __delattr__(self, attr):
        subject = proxy_cache.get(self)
        delattr(subject, attr)

    def __getattribute__(self, attr):
        subject = proxy_cache.get(self)
        if attr == "__subject__":
            return subject
        try:
            return getattr(subject, attr)
        except AttributeError:
            if attr == '__conform__' and type(subject) in _python_types:
                # Conform is an ooddball rejected, but used, way of converting a value
                # Problem is builtins probably don't support it since it was a rejected
                # PEP but the caller may find our Proxy and as it to conform.
                return lambda *a: subject
            raise


def generate_method(name_or_fn):
    fn = None
    if callable(name_or_fn):
        fn = name_or_fn
    if fn is None:
        fn = imported_operators.get(op)
    if fn is None:
        fn = f"__{op.replace('_','')}__"
    return lambda s, *a, **kw: proxied_call(s, fn, *a, **kw)


methods = {}
for op in all_op:
    meth_name = f'__{op.replace("_","")}__'
    if meth_name not in methods:
        methods[meth_name] = generate_method(op)

for op in set(dir(object)) | set(dir(int)):
    if op[0:2] == "__":
        meth_name = f'__{op.replace("_","")}__'
        if meth_name not in methods:
            methods[meth_name] = generate_method(meth_name)

for op in [len, ]:
    methods[f"__{op.__name__}__"] = lambda s, * \
        a, **kw: proxied_call(s, op, *a, **kw)

del methods["__new__"]
del methods["__init__"]
del methods["__getattribute__"]
ComplexProxyMixin = type('ComplexProxy', (object, ), methods)
Proxy = type('Proxy', (BasicProxy, ComplexProxyMixin), {})


def proxy(proxied, config=None, **kw):
    """Proxying can be confusing largely in part that there are lots of edgecases and how you want your proxy to work might vary. In and effort to offer as much configs can be passed into to customize behavior in edge cases. Configs """
    prx = Proxy()
    proxy_cache.proxy(prx, proxied, **kw)
    return prx


def CallbackProxy(cb):
    return proxy(cb, load_once=False)


def LazyProxy(cb):
    return proxy(cb, load_once=True)




import json

import unittest
from mortar.proxy import *

try: long
except NameError: long = int

try: cmp
except NameError: cmp = lambda a, b: (a>b)-(a<b)

class ProxyTestMixin:

    def checkInteger(self, v):
        p = self.proxied(v)
        assert isinstance(v, int)
        assert isinstance(p, int)
        jsn = json.dumps({"key": p})
        self.assertEqual(p|0x10101, v|0x10101)
        self.assertEqual(p&0x10101, v&0x10101)
        self.assertEqual(p^0x10101, v^0x10101)
        self.assertEqual(~p, ~v)
        self.assertEqual(p<<3, v<<3)
        self.assertEqual(p>>2, v>>2)

        self.assertEqual(0x10101|p, 0x10101|v)
        self.assertEqual(0x10101&p, 0x10101&v)
        self.assertEqual(0x10101^p, 0x10101^v)
        self.assertEqual(3<<p, 3<<v)
        self.assertEqual(2>>p, 2>>v)
        for f in hex, oct:
            self.assertEqual(f(p), f(v))

        self.checkNumeric(v)

    def checkNumeric(self, v):
        p = self.proxied(v)
        self.assertEqual(p*2, v*2)
        self.assertEqual(2*p, 2*v)
        self.assertEqual(2**p, 2**v)
        self.assertEqual(p**2, v**2)
        self.assertEqual(-p, -v)
        self.assertEqual(+p, +v)
        for f in abs, int, long, float, hash, complex:
            self.assertEqual(f(p), f(v))

        self.assertEqual(p<22, v<22)
        self.assertEqual(p>=10, v>=10)
        self.assertEqual(p>9.0, v>9.0)
        self.assertEqual(p<=9.25, v<=9.25)
        self.assertEqual(p==7, v==7)
        self.assertEqual(p!=18, v!=18)

        self.assertEqual(16+p, 16+v)
        self.assertEqual(62-p, 62-v)
        self.assertEqual(p+16, v+16)
        self.assertEqual(p-62, v-62)
        self.assertEqual(p%7, v%7)
        self.assertEqual(p/6, v/6)

        self.assertEqual(22<p, 22<v)
        self.assertEqual(10>=p, 10>=v)
        self.assertEqual(9.0>p, 9.0>v)
        self.assertEqual(9.25<=p, 9.25<=v)
        self.assertEqual(7==p, 7==v)
        self.assertEqual(18!=p, 18!=v)

        self.assertEqual(cmp(p,14), cmp(v,14))
        self.assertEqual(cmp(14,p), cmp(14,v))

        self.assertEqual(divmod(p,3), divmod(v,3))

        if v:
            self.assertEqual(divmod(3,p), divmod(3,v))
            self.assertEqual(62//p, 62//v)
            self.assertEqual(7/p, 7/v)
            self.assertEqual(7%p, 7%v)

        self.checkBasics(v)








    def checkList(self, v):
        p = self.proxied(v)
        for i in range(len(v)):
            self.assertEqual(p[i], v[i])
            self.assertEqual(p[i:], v[i:])
            self.assertEqual(p[:i], v[:i])
            self.assertEqual(p[i::-1], v[i::-1])
        self.checkContainer(v)

        c = list(v)
        del p[::1]
        del c[::1]
        self.assertEqual(v, c)

        p[1:1] = [23]
        c[1:1] = [23]
        self.assertEqual(v, c)

    def checkContainer(self, v):
        p = self.proxied(v)
        self.assertEqual(list(p), list(v))
        self.assertEqual(list(iter(p)), list(iter(v)))
        self.assertEqual(len(p), len(v))
        self.assertEqual(42 in p, 42 in v)
        self.assertEqual(99 in p, 99 in v)
        self.checkBasics(v)

    def checkBasics(self, v):
        p = self.proxied(v)
        for f in bool, repr, str:
            self.assertEqual(f(p), f(v))

    def testNumbers(self):
        for i in range(20):
            self.checkInteger(i)

        f = -40
        while f<=20.0:
            self.checkNumeric(f)
            f += 2.25

    def testLists(self):
        try:
            from UserList import UserList
        except ImportError: from collections import UserList
        for d in [1,2], [3,42,59], [99,23,55]:
            self.checkList(d)
            self.checkList(UserList(d))


class InPlaceMixin(ProxyTestMixin):

    def checkInteger(self, vv):
        mk = lambda: (self.proxied(vv), vv)
        p,v = mk(); 
        v|=0x10101; 
        p|=0x10101; 
        self.assertEqual(p.__subject__, v)

        p,v = mk(); p&=0x10101; v&=0x10101; self.assertEqual(p.__subject__, v)
        p,v = mk(); p^=0x10101; v^=0x10101; self.assertEqual(p.__subject__, v)
        p,v = mk(); p<<=3; v<<=3; self.assertEqual(p.__subject__, v)
        p,v = mk(); p>>=3; v>>=3; self.assertEqual(p.__subject__, v)
        ProxyTestMixin.checkInteger(self, vv)

    def checkNumeric(self, vv):
        mk = lambda: (self.proxied(vv), vv)
        p,v = mk(); p+=17; v+=17; self.assertEqual(p.__subject__, v)
        p,v = mk(); p-=22; v-=22; self.assertEqual(p.__subject__, v)
        p,v = mk(); p*=15; v*=15; self.assertEqual(p.__subject__, v)
        p,v = mk(); p//=3; v//=3; self.assertEqual(p.__subject__, v)
        p,v = mk(); p**=2; v**=2; self.assertEqual(p.__subject__, v)
        p,v = mk(); p/=61; v/=61; self.assertEqual(p.__subject__, v)
        p,v = mk(); p%=19; v%=19; self.assertEqual(p.__subject__, v)
        ProxyTestMixin.checkNumeric(self, vv)


class TestCallbackProxy(ProxyTestMixin, unittest.TestCase):
    proxied = lambda self, v: CallbackProxy(lambda:v)

#class TestObjectProxy(InPlaceMixin, unittest.TestCase):
#    proxied = ObjectProxy

class TestLazyProxy(InPlaceMixin, unittest.TestCase):
    proxied = lambda self, v: LazyProxy(lambda:v)





from typing import Type, Union, Text, List, get_args, get_origin, Any
from sqlalchemy.sql.sqltypes import NullType, DateTime, Float, JSON, Integer, String
from sqlalchemy.dialects.mysql.types import TIMESTAMP
from .util import Datetime
from dataclasses import dataclass, is_dataclass, fields, Field, field
from sqlalchemy.types import TypeEngine

VARCHAR_LEN = 256

alchamey_type = {
    Text: lambda *a,**kw: String(length=VARCHAR_LEN),
    List: String(length=VARCHAR_LEN),
    Datetime:DateTime,
    float: Float,
    int: Integer,
    dict: JSON
}


def python_to_alechamy(a_type):
    generic_type = get_origin(a_type)
    if generic_type:
        if generic_type in (dict, list):
            return JSON
        template = list(get_args(a_type))
        if generic_type == Union:
            try:
                template.remove(type(None))
            except ValueError:
                ...  # no worries if it isn't in there.
            if len(template) == 1:
                return python_to_alechamy(template[0])
        else:
            a_type = generic_type
    try:
        if issubclass(a_type, TypeEngine) or is_dataclass(a_type):
            return a_type
    except TypeError:
        ...

    return alchamey_type.get(a_type, String(length=VARCHAR_LEN))


import os
from dataclasses import dataclass, field
from typing import Text

from sqlalchemy import Column, Integer, ForeignKey
from sqlalchemy.orm import relationship

from mortar import persist, metadata


@persist
@dataclass
class Driver(object):
    driver_id: int = field(metadata=Column(Integer, primary_key=True), default=None)
    Name: Text = field(default=None)
    age: int = field(default=None)


@persist
@dataclass
class Car(object):
    Id: int = field(metadata=Column(Integer, primary_key=True))
    Name: Text
    Price: int
    driver: Driver = field(default=None)


if __name__ == "__main__":
    try:
        os.remove('cars.sqlite')
    except:  ...
    metadata.engine('sqlite:///cars.sqlite')

    Car.truncate()
    Driver.truncate()
    dave = Driver(1, 'Dave', 34)
    lnone_car = Car(Id=1, Name='Audi', Price=52642)
    lnone_car.driver = dave
    print(lnone_car)
    metadata.add_all(
        [lnone_car,
            Car(Id=2, Name='Mercedes', Price=57127),
            Car(Id=3, Name='Skoda', Price=9000, driver=dave),
            Car(Id=4, Name='Volvo', Price=29000, driver=dave),
            Car(Id=5, Name='Bentley', Price=350000, driver=dave),
            Car(Id=6, Name='Citroen', Price=21000),
            Car(Id=7, Name='Hummer', Price=41400),
            Car(Id=8, Name='Volkswagen', Price=21600, driver=dave)])

    rs = metadata.query(Car).all()

    for car in rs:
        print(car.driver)

    metadata(Car).delete()
